/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###MODULE###
Name:      Render
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2003-2004.  All rights reserved.
Date:      July 2004
Version:   1.0
Status:    Stable

###DESCRIPTION###

###END###

*****************************************************************************/

//#define DBG_DRAW
//#define DBG_DRAW_ROUTINES /* Use this if you want to debug any outside code subscribing to render drawing routines */
//#define FASTACCESS
//#define SEMAPHORES /* Not recommended, complex ARF_READ / ARF_WRITE operations causes random crashes & hangs (test by rapidly clicking on the file manager menubar) */
//#define SAFE_SEMAPHORES

#define LOCAL_RENDER

#ifdef __linux__
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <errno.h>
#endif

#define __system__
#include <pandora/system/all.h>
#include <pandora/input/joyport.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/picture.h>
#include <pandora/graphics/render.h>
#include <pandora/graphics/screen.h>
#include <pandora/data/config.h>

#include <pandora/modules/render.h>
#include <pandora/modules/screen.h>
#include <pandora/modules/bitmap.h>

#ifdef __linux__
#include <pandora/files/file.h>
#include <pandora/modules/filesystem.h>
#endif

#undef NULL
#define NULL 0

#include <pandora/functions/strtoint.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);
static ERROR CMDOpen(OBJECTPTR);

#define VER_RENDER ((FLOAT)1.0)

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, CMDOpen, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_RENDER, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "July 2004",
   "Render"
};

#define SIZE_FOCUSLIST    30
#define SIZE_RENDERLIST   500 /* Maximum number of render objects allowed in memory */

static struct KernelBase *KernelBase;
static struct ScreenBase *ScreenBase;
static struct BitmapBase *BitmapBase;
static struct RenderList *glRenderList = NULL;
static struct Screen *glScreen = NULL;
static struct Bitmap *glVideoBitmap = NULL;
static struct Bitmap *glComposite = NULL;
static OBJECTPTR RenderClass = NULL;
static OBJECTPTR modScreen = NULL;
static OBJECTPTR modBitmap = NULL;
static OBJECTPTR modRender = NULL;
static LONG glMaxBitmapWait = 0;
static WORD glNoDrawing = 0;
static WORD glVideoCount = 0;
static WORD glNoExpose = 0;
static OBJECTID glKeyboardID = 0;
static OBJECTID glBroadcastID = 0;
static OBJECTID glFreeExpose = 0;
static BYTE glSlowVideoBlit = -1;
static WORD glVolatileIndex = 0;
static UBYTE glIgnoreCursor = FALSE;

enum {
   DRAG_NONE=0,
   DRAG_ANCHOR,
   DRAG_NORMAL
};

enum {
   STAGE_PRECOPY=1,
   STAGE_AFTERCOPY,
   STAGE_POSTCOMPOSITE
};

#ifdef SEMAPHORES
static LONG glListSemaphore = 0;
#else
static UBYTE glListCount = 0;
#endif

#ifdef DBG_DRAW
static BYTE glDebugExposure = FALSE;
#endif

/*****************************************************************************
** Module function list.
*/

static struct FunctionField argsGetRenderInfo[]          = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Info", ARG_PTR }, { "InfoSize", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsLockDrawableBitmap[]     = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Bitmap", ARG_PTR }, { "Info", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsUnlockDrawableBitmap[]   = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Bitmap", ARG_PTR }, { NULL, NULL } };
static struct FunctionField argsExposeDrawable[]         = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsBlitDrawable[]           = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Bitmap", ARG_PTR }, { "Flags", ARG_LONG }, { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { "XDest", ARG_LONG }, { "YDest", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsAccessDrawableSurface[]  = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Surface", ARG_PTRRESULT }, { "Size", ARG_PTRSIZE }, { NULL, NULL } };
static struct FunctionField argsReleaseDrawableSurface[] = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Surface", ARG_PTRRESULT }, { NULL, NULL } };
static struct FunctionField argsAccessRenderList[]       = { { "List", ARG_PTR }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsReleaseRenderList[]      = { { "Void", ARG_VOID }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSetDrawableRegion[]      = { { "Error", ARG_ERROR }, { "Object", ARG_LONG }, { "Region", ARG_PTR }, { NULL, NULL } };
static struct FunctionField argsClippedCallback[]        = { { "Error", ARG_ERROR }, { "Object", ARG_PTR },  { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsGetUserFocus[]           = { { "Drawable", ARG_LONG }, { NULL, NULL } };

static struct Function JumpTableV1[] = {
   { GetRenderInfo,          "GetRenderInfo",          argsGetRenderInfo },
   { LockDrawableBitmap,     "LockDrawableBitmap",     argsLockDrawableBitmap },
   { UnlockDrawableBitmap,   "UnlockDrawableBitmap",   argsUnlockDrawableBitmap },
   { ExposeDrawable,         "ExposeDrawable",         argsExposeDrawable },
   { BlitDrawable,           "BlitDrawable",           argsBlitDrawable },
   { AccessDrawableSurface,  "AccessDrawableSurface",  argsAccessDrawableSurface },
   { ReleaseDrawableSurface, "ReleaseDrawableSurface", argsReleaseDrawableSurface },
   { AccessRenderList,       "AccessRenderList",       argsAccessRenderList },
   { ReleaseRenderList,      "ReleaseRenderList",      argsReleaseRenderList },
   { SetDrawableRegion,      "SetDrawableRegion",      argsSetDrawableRegion },
   { ClippedCallback,        "ClippedCallback",        argsClippedCallback },
   { GetUserFocus,           "GetUserFocus",           argsGetUserFocus },
   { ForbidExpose,           "ForbidExpose",           NULL },
   { PermitExpose,           "PermitExpose",           NULL },
   { NULL, NULL }
};

/****************************************************************************/

#ifdef _WIN32
APTR winGetDC(APTR Window);
void winReleaseDC(APTR Window, APTR DC);
void winSetRenderID(APTR Window, LONG RenderID);
#endif

#ifdef FASTACCESS
static void  FastCopyBackground(struct RenderList *, WORD, WORD, WORD, WORD, WORD, WORD, WORD, struct Bitmap *, OBJECTID, WORD, BYTE);
#else
static void  CopyBackground(struct RenderList *, WORD, WORD, WORD, WORD, WORD, WORD, WORD, struct Bitmap *, struct Bitmap *, WORD, BYTE);
#endif

static struct Bitmap * AccessVideo(OBJECTID);
static UBYTE CheckVisibility(struct RenderList *, WORD);
static UBYTE CheckVolatile(struct RenderList *, WORD);
static ERROR CreateRenderClass(void);
static void  ExposeBuffer(struct RenderList *, WORD, WORD, LONG, LONG, LONG, LONG, struct Bitmap *, struct Bitmap *);
static WORD  FindBitmapOwner(struct RenderList *, WORD);
static ERROR RedrawDrawable(OBJECTID, LONG, LONG, LONG, LONG, LONG);
static void InvalidateExposedOverlap(struct Render *, struct RenderList *, LONG, LONG, LONG, LONG, LONG, LONG, struct Bitmap *);
static ERROR MoveGraphics(struct Render *, struct RenderList *, WORD, LONG, LONG, LONG, LONG, LONG, LONG, LONG, LONG, LONG, LONG, LONG);
static void  MoveLayer(struct Render *, LONG, LONG);
static LONG  MsgHandler(LONG, LONG, APTR, LONG);
static void  PrepareBackground(struct Render *, struct RenderList *, WORD, struct Bitmap *, struct ClipRectangle *, BYTE);
static void RedrawNonIntersect(OBJECTID, struct ClipRectangle *, struct ClipRectangle *, LONG, LONG);
static void  ReleaseVideo();
static ERROR TrackLayer(struct Render *);
static void  UntrackLayer(OBJECTID);
static ERROR UpdateRenderList(struct Render *);

#define IRF_IGNORE_NVCHILDREN 0x00000001
#define IRF_IGNORE_CHILDREN   0x00000002
#define IRF_SINGLE_BITMAP     0x00000004
#define IRF_RELATIVE          0x00000008
//#define IRF_EXPOSE            0x00000010

#define URF_HATE_CHILDREN     0x00000001

/****************************************************************************/

static UBYTE SlowVideoBlit(OBJECTID ScreenID)
{
   SCREENINFO info;

   if (glSlowVideoBlit IS -1) {
      glSlowVideoBlit = FALSE;
      if (GetScreenInfo(ScreenID, &info, sizeof(info)) IS ERR_Okay) {
         if (!(info.AccelFlags & ACF_VIDEOBLIT)) glSlowVideoBlit = TRUE;
      }
   }

   return(glSlowVideoBlit);
}

static inline WORD FindDrawableIndex(struct RenderList *List, OBJECTID DrawableID)
{
   WORD i;
   for (i=0; List[i].ObjectID; i++) if (List[i].ObjectID IS DrawableID) return(i);
   return(-1);
}

static inline void ClipRectangle(struct ClipRectangle *rect, struct ClipRectangle *clip)
{
   if (rect->Left   < clip->Left)   rect->left   = clip->Left;
   if (rect->Top    < clip->Top)    rect->top    = clip->Top;
   if (rect->Right  > clip->Right)  rect->right  = clip->Right;
   if (rect->Bottom > clip->Bottom) rect->bottom = clip->Bottom;
}

/*****************************************************************************
** Internal: test_x11_socket()
*/

#ifdef __linux__
static LONG test_x11(STRING Path)
{
   struct sockaddr_un sockname;
   LONG namelen, fd, err;
   WORD i;

   if ((fd = socket(PF_UNIX, SOCK_STREAM, 0)) != -1) {
      fcntl(fd, F_SETFL, O_NONBLOCK);

      sockname.sun_family = AF_UNIX;
      ClearMemory(sockname.sun_path, sizeof(sockname.sun_path));
      for (i=0; i < sizeof(sockname.sun_path) AND (Path[i]); i++) sockname.sun_path[i] = Path[i];
      namelen = i + sizeof(sockname.sun_family);

      if (connect(fd, (struct sockaddr *)&sockname, namelen) < 0) {
         DPrintF("TestX11Socket:","Socket %s failed: %s", Path, strerror(errno));
         err = errno;
         close(fd);

         /* If the error was ENOENT, the server may be starting up
         ** and we should try again.
         **
         ** If the error was EWOULDBLOCK or EINPROGRESS then the socket
         ** was non-blocking and we should poll using select
         **
         ** If the error was EINTR, the connect was interrupted and we
         ** should try again.
         */

         if ((err IS EINTR)) return(ERR_Okay);
         else if ((err IS EWOULDBLOCK) OR (err IS EINPROGRESS)) return(ERR_Okay);
         else {
            DPrintF("TestX11:","Connect Error: %s", strerror(err));
            return(ERR_Failed);
         }
      }
      else DPrintF("TestX11Socket:","Connected to %s", Path);

      close(fd);

      return(ERR_Okay);
   }
   else DPrintF("TestSocket:","Failed to open a socket.");

   return(ERR_Failed);
}
#endif

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTPTR object;
   OBJECTID object_id;
   MEMORYID MemoryID;
   ERROR error;
   BYTE display;
   enum {
      DISPLAY_X11=1,
      DISPLAY_NATIVE
   };

   KernelBase = argKernelBase;

   display = 0;

#ifdef __linux__
   /* Check if X11 is running.  This will affect the screen and bitmap drivers that
   ** we want to load.  If X11 is not running, we'll use native graphics mode.
   */

   STRING str;
   APTR address;

   /*** Check if SNAP is already open ***/

   if (AccessMemory(RPM_SNAPExports, MEM_READ, &address) IS ERR_Okay) {
      DPrintF("InitModule:","[Render] Detected SNAP as running.");
      display = DISPLAY_NATIVE;
      ReleaseMemory(address);
   }

   if (!display) {
      /*** Check user commandline arguments for --gfxdriver **/

      STRING *args;
      WORD i;
      if ((GetField(CurrentTask(), FID_ArgsList, FT_POINTER, &args) IS ERR_Okay) AND (args)) {
         for (i=0; args[i]; i++) {
            if (StrCompare(args[i], "--gfxdriver=native", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               display = DISPLAY_NATIVE;
            }
            else if (StrCompare(args[i], "--gfxdriver=x11", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               display = DISPLAY_X11;
            }
         }
      }
   }

   if (!display) {
      /*** Check if the user has forced a display preference in his display.cfg file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &object, NULL,
         FID_Location|TSTRING, "user:config/display.cfg",
         TAGEND) IS ERR_Okay) {

         if (cfgRead(object, "SCREEN", "Driver", &str) IS ERR_Okay) {
            if (StrCompare("Native", str, 0, STR_MATCHLENGTH) IS ERR_Okay)    display = DISPLAY_NATIVE;
            else if (StrCompare("VGA", str, 0, STR_MATCHLENGTH) IS ERR_Okay)  display = DISPLAY_NATIVE;
            else if (StrCompare("VESA", str, 0, STR_MATCHLENGTH) IS ERR_Okay) display = DISPLAY_NATIVE;
            else if (StrCompare("SNAP", str, 0, STR_MATCHLENGTH) IS ERR_Okay) display = DISPLAY_NATIVE;
            else if (StrCompare("X11", str, 0, STR_MATCHLENGTH) IS ERR_Okay)  display = DISPLAY_X11;
         }

         acFree(object);
      }
   }

   if (!display) {
      /*** Check if X11 is running by scanning /tmp/.X11-unix ***/

      WORD i, j;
      UBYTE buffer[80] = "/tmp/.X11-unix/";
      STRING x11[] = { "X", "X0", "X1", "X2", "X3", "X5", "X6", "X7", "X8", "X9", "X10" };

      for (i=0; (i < ARRAYSIZE(x11)); i++) {
         for (j=0; x11[i][j]; j++) buffer[15+j] = x11[i][j];
         buffer[15+j] = 0;
         if (test_x11(buffer) IS ERR_Okay) {
            display = DISPLAY_X11;
            break;
         }
      }

      /*** Check if X11 is running according to the process list ***/
   }

   if (!display) {
      DPrintF("InitModule:","Using default display driver: Native");
      display = DISPLAY_NATIVE;
   }
   else DPrintF("InitModule:","[Render] Display will be: %s", (display IS DISPLAY_NATIVE) ? "Native" : "X11");

#else
   display = DISPLAY_NATIVE;
#endif

   /*** Add a message handler to the system for responding to interface messages ***/

   if (AddMsgHandler(&MsgHandler) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_Failed));
   }

   /*** Allocate the FocusList memory block ***/

   MemoryID = RPM_FocusList;
   error = AllocMemory((SIZE_FOCUSLIST) * sizeof(OBJECTID), MEM_UNTRACKED|MEM_RESERVED|MEM_PUBLIC, NULL, &MemoryID);
   if ((error != ERR_Okay) AND (error != ERR_ResourceExists)) {
      return(ObjectError(argModule, ERH_InitModule, ERR_AllocMemory));
   }

   GetField(argModule, FID_Master, FT_POINTER, &modRender);

   /*** Allocate the RenderList memory block and its associated semaphore ***/

#ifdef SEMAPHORES

   MemoryID = RPM_RenderList;
   error = AllocMemory(SIZE_RENDERLIST * sizeof(struct RenderList), MEM_UNTRACKED|MEM_RESERVED|MEM_PUBLIC|MEM_NOBLOCKING, &glRenderList, &MemoryID);
   if (error IS ERR_ResourceExists) {
      if (AccessMemory(RPM_RenderList, MEM_READWRITE, &glRenderList) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_AccessMemory));
      }
   }
   else if (error != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_AllocMemory));
   }

   if (AllocSemaphore("renderlist", 100, NULL, &glListSemaphore) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_AllocSemaphore));
   }

#else

   MemoryID = RPM_RenderList;
   error = AllocMemory(SIZE_RENDERLIST * sizeof(struct RenderList), MEM_UNTRACKED|MEM_RESERVED|MEM_PUBLIC, NULL, &MemoryID);
   if ((error != ERR_Okay) AND (error != ERR_ResourceExists)) {
      return(ObjectError(argModule, ERH_InitModule, ERR_AllocMemory));
   }

#endif

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modScreen, NULL,
      FID_Name|TSTRING,   (display IS DISPLAY_NATIVE) ? "screen" : "classes:graphics/screen-x11",
      FID_Version|TFLOAT, MODVERSION_SCREEN,
      TAGEND) IS ERR_Okay) {
      if (GetField(modScreen, FID_ModBase, FT_POINTER, &ScreenBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modBitmap, NULL,
      FID_Name|TSTRING,   "bitmap",
      FID_Version|TFLOAT, MODVERSION_BITMAP,
      TAGEND) IS ERR_Okay) {
      if (GetField(modBitmap, FID_ModBase, FT_POINTER, &BitmapBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /* It would make sense to load the pointer module in here, but it causes recursive
   ** issues (the Render module requires the Pointer module, the Pointer module requires
   ** the Render module).  Resolving this problem would help, of course.
   */

   /*** Find system objects ***/

   FastFindObject("SystemBroadcaster", ID_BROADCAST, &glBroadcastID, 1, NULL);

   if (CheckObjectExists(NULL, "SystemKeyboard") != ERR_Okay) {
      if (CreateObject(ID_KEYBOARD, NF_PUBLIC|NF_UNTRACKED|NF_UNIQUE, &object, &object_id,
            FID_Name|TSTRING, "SystemKeyboard",
            TAGEND) IS ERR_Okay) {
         acActivate(object);
         ReleaseObject(object);
      }
   }

   FastFindObject("SystemKeyboard", ID_KEYBOARD, &glKeyboardID, 1, NULL);

   return(CreateRenderClass());
}

/*****************************************************************************
** Command: Open()
*/

static ERROR CMDOpen(OBJECTPTR Module)
{
   SetField(Module, FID_FunctionList, FT_POINTER, JumpTableV1);
   return(ERR_Okay);
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   RemoveMsgHandler(MsgHandler);

   if (glComposite) { acFree(glComposite); glComposite = NULL; }

   #ifdef SEMAPHORES
      if (glRenderList) { ReleaseMemory(glRenderList); glRenderList = NULL; }
      if (glListSemaphore) { FreeSemaphore(glListSemaphore); glListSemaphore = NULL; }
   #endif

   if (modBitmap)   { acFree(modBitmap); modBitmap = NULL; }
   if (modScreen)   { acFree(modScreen); modScreen = NULL; }
   if (RenderClass) { acFree(RenderClass); RenderClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: MsgHandler()
** Short:    Handles incoming interface messages.
*/

static LONG MsgHandler(LONG UniqueID, LONG Type, APTR Data, LONG Size)
{
   if ((Type IS MSGID_EXPOSE) AND (Data) AND (Size >= sizeof(struct ExposeMessage))) {
      struct ExposeMessage *expose;

      expose = Data;

      glMaxBitmapWait = 2000;
      ExposeDrawable(expose->ObjectID, expose->XCoord, expose->YCoord, expose->Width, expose->Height, expose->Flags);
      glMaxBitmapWait = 0;
      return(TRUE);
   }
   else return(FALSE);
}

/*****************************************************************************

###FUNCTION###
Name:     AccessDrawableSurface
Short:    Provides free access to bitmap data without performing any locks.
Synopsis: ERROR AccessDrawableSurface(OBJECTID RenderID, struct BitmapSurface *Surface, LONG Size)

###DESCRIPTION###
<p>The AccessDrawableSurface() function is used to gain free access to the bitmap
data of any given render object.  'Free access' means that the graphics data
will not be locked or synchronised while you have access to it, effectively
allowing other tasks to also use the bitmap at the same time.  This can be
extremely useful for unsynchronised, high speed read/write access.  However,
the lack of synchronisation means that if multiple tasks are using the bitmap
at once, you may encounter inconsistent results.</p>

<p>The surface information is returned in a BitmapSurface structure.  More
information on this structure is available in the documentation for the
Bitmap module's BlitSurface() function.</p>

<p>After you have finished with the bitmap, you must match this function with
a call to ReleaseDrawableSurface.  Multiple calls to AccessDrawableSurface will
nest.  Never re-use the BitmapSurface structure when making calls to this
function.</p>

<p>If you would like your bitmap access to be synchronised, use the
LockDrawableBitmap() function instead.</p>

###INPUT###
RenderID: Reference to the drawable ID that you want to access.
Surface: Pointer to a BitmapSurface structure that will receive the surface information.
Size: Size of the BitmapSurface structure.

###RESULT###
ERR_Okay: Access successful.
ERR_Args: Invalid arguments were supplied.
ERR_Search: The supplied RenderID did not refer to a recognised render object.
ERR_AccessMemory: Failed to access the internal renderlist memory structure, or the bitmap data was not accessible.

###SEE ALSO###
Function: ReleaseDrawableSurface, LockDrawableBitmap

###END###

*****************************************************************************/

static ERROR AccessDrawableSurface(OBJECTID RenderID, struct BitmapSurface *Surface, LONG Size)
{
   struct RenderList *list;
   struct Bitmap *bitmap;
   SCREENINFO scrinfo;
   WORD i, root;

   if ((!RenderID) OR (!Surface) OR (Size < 4)) {
      return(SendError(ERH_AccessSurface, ERR_Args));
   }

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS RenderID) {
            root = FindBitmapOwner(list, i);

            /*** Set the surface version ***/

            if (Size >= sizeof(struct BitmapSurfaceV2)) Surface->Version = 2;
            else Surface->Version = 1;

            if (!list[root].DataMID) {
               /* If a DataMID is not available, the render surface is held in video
               ** memory.  The only way to legally acquire video memory addresses is
               ** to access the bitmap object directly.
               */
               if (Surface->Version IS 1) {
                  ReleaseRenderList(ARF_READ);
                  return(ERR_NoSupport);
               }
               else if (AccessObject(list[root].BitmapID, 2000, &bitmap) IS ERR_Okay) {
                  if (bitmap->Data) {
                     Surface->Data          = bitmap->Data;
                     Surface->Width         = list[root].Width;
                     Surface->ByteWidth     = list[root].LineWidth;
                     Surface->Height        = list[i].Height;
                     Surface->BitsPerPixel  = list[root].BitsPerPixel;
                     Surface->BytesPerPixel = list[root].BytesPerPixel;
                     Surface->XOffset       = list[i].AbsX - list[root].AbsX;
                     Surface->YOffset       = list[i].AbsY - list[root].AbsY;
                     Surface->ClipLeft      = 0;
                     Surface->ClipTop       = 0;
                     Surface->ClipRight     = list[i].Width;
                     Surface->ClipBottom    = list[i].Height;
                     Surface->Private       = bitmap;

                     if (GetScreenInfo(NULL, &scrinfo, sizeof(scrinfo)) IS ERR_Okay) {
                        CopyMemory(&scrinfo.PixelFormat, &Surface->Format, sizeof(Surface->Format));
                     }

                     ReleaseRenderList(ARF_READ);
                     return(ERR_Okay);
                  }
                  else {
                     ReleaseObject(bitmap);
                     ReleaseRenderList(ARF_READ);
                     return(SendError(ERH_AccessSurface, ERR_AccessMemory));
                  }
               }
               else {
                  ReleaseRenderList(ARF_READ);
                  return(SendError(ERH_AccessSurface, ERR_AccessObject));
               }
            }
            else if (AccessMemory(list[root].DataMID, MEM_READ|MEM_WRITE|MEM_NOBLOCKING, &Surface->Data) IS ERR_Okay) {
               Surface->Width         = list[root].Width;
               Surface->ByteWidth     = list[root].LineWidth;
               Surface->Height        = list[i].Height;
               Surface->BitsPerPixel  = list[root].BitsPerPixel;
               Surface->BytesPerPixel = list[root].BytesPerPixel;
               Surface->XOffset       = list[i].AbsX - list[root].AbsX;
               Surface->YOffset       = list[i].AbsY - list[root].AbsY;
               Surface->ClipLeft      = 0;
               Surface->ClipTop       = 0;
               Surface->ClipRight     = list[i].Width;
               Surface->ClipBottom    = list[i].Height;
               if (Surface->Version IS 2) Surface->Private = NULL;

               if (GetScreenInfo(NULL, &scrinfo, sizeof(scrinfo)) IS ERR_Okay) {
                  CopyMemory(&scrinfo.PixelFormat, &Surface->Format, sizeof(Surface->Format));
               }

               ReleaseRenderList(ARF_READ);
               return(ERR_Okay);
            }
            else {
               DPrintF("@AccessSurface()","Failed to access memblock #%d of render object #%d", list[root].DataMID, RenderID);
               ReleaseRenderList(ARF_READ);
               return(ERR_AccessMemory);
            }
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ERR_Search);
   }
   else return(SendError(ERH_AccessSurface, ERR_AccessMemory));
}

/*****************************************************************************

###PRIVATE###
Name:     AccessRenderList()
Short:    Grants access to the internal renderlist array.
Synopsis: struct RenderList * AccessRenderList(LONG Flags, LONG *Size)

###DESCRIPTION###

###INPUT###
Flags: Specify ARF_WRITE if writing to the list, otherwise ARF_READ must be set.
Size: The size of each list entry will be returned in the 32-bit integer referred to in this argument.

###SEE ALSO###
Function: ReleaseRenderList

###END###

*****************************************************************************/

static struct RenderList * AccessRenderList(LONG Flags, LONG *Size)
{
#ifdef SEMAPHORES
   LONG Timeout;

#ifdef SAFE_SEMAPHORES
   Flags = ARF_WRITE;
#endif

   if (Size) *Size = sizeof(struct RenderList);

   if (Flags & ARF_NODELAY) Timeout = 20;
   else Timeout = 5000;

   if (Flags & (ARF_WRITE|ARF_UPDATE)) {
      /*** Blocking access to the renderlist semaphore is required ***/
      if (AccessSemaphore(glListSemaphore, Timeout, NULL) IS ERR_Okay) {
         return(glRenderList);
      }
      else DPrintF("@AccessRenderList()","Failed to gain blocking access to the renderlist.");
   }
   else if (Flags & ARF_READ) {
      /*** Non-Blocking access to the renderlist semaphore is required ***/
      if (AccessSemaphore(glListSemaphore, Timeout, SMF_NONBLOCKING) IS ERR_Okay) {
         return(glRenderList);
      }
      else DPrintF("@AccessRenderList()","Failed to gain non-blocking access to the renderlist.");
   }
   else DPrintF("@AccessRenderList()","ARF_READ or ARF_WRITE must be specified.");

   return(NULL);

#else

   if (Size) *Size = sizeof(struct RenderList);

   if (!glRenderList) {
      LONG timeout;
      ERROR error;

      if (Flags & ARF_NODELAY) {
         timeout = SetResource(RES_TIMEOUT, 20);
         error = AccessMemory(RPM_RenderList, MEM_READWRITE, &glRenderList);
         SetResource(RES_TIMEOUT, timeout);
      }
      else error = AccessMemory(RPM_RenderList, MEM_READWRITE, &glRenderList);

      if (error IS ERR_Okay) glListCount = 1;
   }
   else glListCount++;

   return(glRenderList);

#endif
}

/*****************************************************************************

###FUNCTION###
Name:     BlitDrawable
Short:    Copies drawable graphics data into any Bitmap object.
Synopsis: ERROR BlitDrawable(OBJECTID RenderID, struct Bitmap *Bitmap, LONG Flags, LONG XCoord, LONG YCoord, LONG Width, LONG Height, LONG XDest, LONG YDest)

###DESCRIPTION###
<p>This function will copy the graphics data from any drawable object into a
Bitmap of your choosing.  This is the fastest and most convenient way to get
graphics information out of any drawable.  Because Pandora uses graphics
buffering, you are also guaranteed that the resulting graphics will not be
obscured by any overlapping drawables that are on the display.</p>

<p>In the event that the owner of the drawable is drawing to the graphics
buffer at the time that you call this function, the results may be slightly
out of sync.  If this could be a problem, set the BDF_SYNC option in the Flags
parameter.  Keep in mind that syncing has the negative side effect of having to
wait for the other task to complete its draw process, which can potentially
result in time lags.</p>

###INPUT###
RenderID: The ID of the render object that you want to blit from.
Flags: Optional flags.
XCoord: The horizontal source coordinate.
YCoord: The vertical source coordinate.
Width: The width of the graphic that you want to copy.
Height: The height of the graphic that you want to copy.
XDest: The horizontal destination coordinate.
YDest: The vertical destination coordinate.
Bitmap: Pointer to the destination Bitmap object.

###RESULT###
ERR_Okay: The blit was successful.
ERR_Args: Invalid arguments were specified.
ERR_Search: The supplied RenderID did not refer to a recognised render object.
ERR_AccessMemory: Failed to access the internal renderlist memory structure.

###END###

*****************************************************************************/

static ERROR BlitDrawable(OBJECTID RenderID, struct Bitmap *Bitmap, LONG Flags,
          LONG XCoord, LONG YCoord, LONG Width, LONG Height, LONG XDest, LONG YDest)
{
   struct BitmapSurface surface;
   struct Bitmap *src;
   struct RenderList *list;
   WORD i, root;
   LONG flags;
   BYTE composite;

   if ((!RenderID) OR (!Bitmap)) return(SendError(ERH_BlitDrawable, ERR_Args));

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS RenderID) {
            if (XCoord < 0) { XDest -= XCoord; Width  += XCoord; XCoord = 0; }
            if (YCoord < 0) { YDest -= YCoord; Height += YCoord; YCoord = 0; }
            if (Width  > list[i].Width)  Width  = list[i].Width;
            if (Height > list[i].Height) Height = list[i].Height;

            /*** Find the bitmap root ***/

            root = FindBitmapOwner(list, i);

            if ((Flags & BDF_SYNC) OR (!list[root].DataMID)) {
               if (AccessObject(list[root].BitmapID, 4000, &src) IS ERR_Okay) {
                  src->XOffset    = list[i].AbsX - list[root].AbsX;
                  src->YOffset    = list[i].AbsY - list[root].AbsY;
                  src->ClipLeft   = 0;
                  src->ClipTop    = 0;
                  src->ClipRight  = src->Width;
                  src->ClipBottom = src->Height;
                  if (list[i].Flags & RNF_POSTCOMPOSITE) composite = TRUE;
                  else composite = FALSE;

                  ReleaseRenderList(ARF_READ);

                  if (composite) {
                     flags = src->Flags;
                     src->Flags |= BMF_ALPHABLEND;
                     ActionTags(MT_BlitArea, src, Bitmap, NULL, XCoord, YCoord, Width, Height, XDest, YDest);
                     src->Flags = flags;
                  }
                  else ActionTags(MT_BlitArea, src, Bitmap, NULL, XCoord, YCoord, Width, Height, XDest, YDest);

                  ReleaseObject(src);
                  return(ERR_Okay);
               }
               else {
                  ReleaseRenderList(ARF_READ);
                  return(SendError(ERH_BlitDrawable, ERR_AccessObject));
               }
            }
            else if (AccessMemory(list[root].DataMID, MEM_READ, &surface.Data) IS ERR_Okay) {
               surface.XOffset       = list[i].AbsX - list[root].AbsX;
               surface.YOffset       = list[i].AbsY - list[root].AbsY;
               surface.ByteWidth     = list[root].LineWidth;
               surface.Height        = list[i].Height;
               surface.BitsPerPixel  = list[root].BitsPerPixel;
               surface.BytesPerPixel = list[root].BytesPerPixel;
               if (list[i].Flags & RNF_POSTCOMPOSITE) composite = TRUE;
               else composite = FALSE;
               ReleaseRenderList(ARF_READ);

               if (composite) {
                  BlitSurface(&surface, Bitmap, BSF_DEFAULTFORMAT|BSF_OFFSET|BSF_ALPHA, XCoord, YCoord, Width, Height, XDest, YDest);
               }
               else BlitSurface(&surface, Bitmap, BSF_DEFAULTFORMAT|BSF_OFFSET, XCoord, YCoord, Width, Height, XDest, YDest);

               ReleaseMemory(surface.Data);
               return(ERR_Okay);
            }
            else return(SendError(ERH_BlitDrawable, ERR_AccessMemory));
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ERR_Search);
   }
   else return(SendError(ERH_BlitDrawable, ERR_AccessMemory));

   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     ExposeDrawable()
Synopsis: ERROR ExposeDrawable(OBJECTID DrawableID, LONG XCoord, LONG YCoord, LONG Width, LONG Height, LONG Flags)
Short:    Exposes any drawable region for which you have an ID.

###DESCRIPTION###
<p>This expose routine will expose everything within a drawable area.  This
will include all child drawables that intersect with the region being exposed
if you set the EXF_CHILDREN flag.</p>

###INPUT###
DrawableID: The ID of the render object that you want to expose.
XCoord: The horizontal coordinate of the area to expose.
YCoord: The vertical coordinate of the area to expose.
Width: The width of the expose area.
Height: The height of the expose area.
Flags: Optional flags - EXF_CHILDREN will expose all intersecting child regions.

###RESULT###
ERR_Okay: The function completed successfully.
ERR_Search: The RenderID does not refer to an existing drawable object.
ERR_AccessMemory: The internal renderlist could not be accessed.

###END###

*****************************************************************************/

static ERROR ExposeDrawable(OBJECTID RenderID, LONG XCoord, LONG YCoord, LONG Width, LONG Height, LONG Flags)
{
   struct RenderList *list;
   struct Bitmap *video, *bitmap;
   struct ClipRectangle abs, childexpose;
   struct mtExpose expose;
   WORD i, j, index, level;
   UBYTE skip;
   OBJECTID parent_id, id;

   if (glNoDrawing) return(ERR_Okay);

   if ((Width < 1) OR (Height < 1)) return(ERR_Okay);

   video = NULL;

   if (!(list = AccessRenderList(ARF_READ, NULL))) {
      DPrintF("@ExposeDrawable()","Unable to access the renderlist.");
      return(ERR_AccessMemory);
   }

   if ((index = FindDrawableIndex(list, RenderID)) IS -1) {
      ReleaseRenderList(ARF_READ);
      DPrintF("@ExposeDrawable()","[%d] Drawable not found in the renderlist.", RenderID);
      return(ERR_Search);
   }

   if ((!(list[index].Flags & RNF_VISIBLE)) OR ((list[index].Width < 1) OR (list[index].Height < 1))) {
      ReleaseRenderList(ARF_READ);
      #ifdef DBG_DRAW
         DPrintF("ExposeDrawable()","[%d] Object invisible or too small to draw.", RenderID);
      #endif
      return(ERR_Okay);
   }

   /*** Calculate the absolute coordinates of the exposed area ***/

   if (Flags & EXF_ABSOLUTE) {
      abs.left   = XCoord;
      abs.top    = YCoord;
      abs.right  = Width;
      abs.bottom = Height;
      Flags &= ~EXF_ABSOLUTE;
   }
   else {
      abs.left   = list[index].AbsX + XCoord;
      abs.top    = list[index].AbsY + YCoord;
      abs.right  = abs.left + Width;
      abs.bottom = abs.top  + Height;
   }

   #ifdef DBG_DRAW
      DPrintF("~ExposeDrawable()","[Render:%d] %dx%d,%dx%d Flags: $%.4x", RenderID, abs.left, abs.top, abs.right-abs.left, abs.bottom-abs.top, Flags);
   #endif

   /*** If the object is invisible, we need to scan back to a visible parent ***/

   if (list[index].Flags & (RNF_INVISIBLE|RNF_REGION)) {
      #ifdef DBG_DRAW
         DPrintF("ExposeDrawable:","Drawable is %s; scan to solid starting from index %d.", (list[index].Flags & RNF_REGION) ? "a region" : "invisible", index);
      #endif

      id = list[index].ObjectID;
      for (j=index; j > 0; j--) {
         if (list[j].ObjectID != id) continue;
         if (list[j].Flags & (RNF_INVISIBLE|RNF_REGION)) id = list[j].ParentID;
         else break;
      }
      Flags |= EXF_CHILDREN;
      index = j;

      #ifdef DBG_DRAW
         DPrintF("ExposeDrawable:","New index %d.", index);
      #endif
   }

   /* Check if the exposed dimensions are outside of our boundary and/or
   ** our parent(s) boundaries.  If so then we must restrict the exposed
   ** dimensions.
   **
   ** NOTE: This loop looks strange but is correct & fast, so don't touch.
   */

   for (i=index, parent_id = RenderID; ;) {
      if (!(list[i].Flags & RNF_VISIBLE)) goto exit;
      ClipRectangle(&abs, (struct ClipRectangle *)&list[i].Left);
      if (!(parent_id = list[i].ParentID)) break;
      i--;
      while (list[i].ObjectID != parent_id) i--;
   }

   if ((abs.left >= abs.right) OR (abs.top >= abs.bottom)) goto exit;

   /*** Check that the expose area actually overlaps the target drawable ***/

   if (abs.left   >= list[index].Right) goto exit;
   if (abs.top    >= list[index].Bottom) goto exit;
   if (abs.right  <= list[index].Left) goto exit;
   if (abs.bottom <= list[index].Top) goto exit;

   /* Cursor split routine.  The purpose of this is to eliminate as much flicker
   ** as possible from the cursor when exposing large areas.
   **
   ** We scan for the software cursor to see if the bottom of the cursor
   ** intersects with our expose region.  If it does, split ExposeDrawable()
   ** into top and bottom regions.
   */

   if (!(Flags & EXF_CURSOR_SPLIT)) {
      WORD cursor;

      for (cursor=index+1; (list[cursor].ObjectID) AND (!(list[cursor].Flags & RNF_CURSOR)); cursor++);

      if ((list[cursor].ObjectID) AND (list[cursor].Bottom < abs.bottom) AND (list[cursor].Bottom > abs.top) AND
          (list[cursor].Right > abs.left) AND (list[cursor].Left < abs.right)) {
         #ifdef DBG_DRAW
            DPrintF("~ExposeDrawable:","Splitting cursor.");
         #endif

         ExposeDrawable(RenderID, abs.left, abs.top, abs.right, list[cursor].Bottom, EXF_CURSOR_SPLIT|EXF_ABSOLUTE|Flags);
         ExposeDrawable(RenderID, abs.left, list[cursor].Bottom, abs.right, abs.bottom, EXF_CURSOR_SPLIT|EXF_ABSOLUTE|Flags);

         #ifdef DBG_DRAW
            StepBack();
         #endif

         goto exit;
      }
   }


   /* The expose routine starts from the front and works to the back, so
   ** if the EXF_CHILDREN flag has been specified, the first thing we do is
   ** scan to the final child that is listed in this particular area.
   */

   if (Flags & EXF_CHILDREN) {
      /*** Change the index to the root bitmap of the exposed object ***/
      index = FindBitmapOwner(list, index);
      for (i=index; list[i+1].Level > list[index].Level; i++); /* Go all the way to the end of the list */
   }
   else i = index;

   for (; i >= index; i--) {
      /*** Ignore regions and non-visible drawables ***/

      if (list[i].Flags & (RNF_REGION|RNF_INVISIBLE)) continue;

      if ((list[i].Flags & RNF_CURSOR) AND (list[i].ObjectID != RenderID)) continue;

      /*** If this is not a root bitmap object, skip it (i.e. consider it like a region) ***/

      skip = FALSE;
      parent_id = list[i].ParentID;
      for (j=i-1; j >= index; j--) {
         if (list[j].ObjectID IS parent_id) {
            if (list[j].BitmapID IS list[i].BitmapID) skip = TRUE;
            break;
         }
      }
      if (skip) continue;

      childexpose = abs;

      if (i != index) {
         /*** Check this child object and its parents to make sure they are visible ***/

         parent_id = list[i].ObjectID;
         for (j=i; (j >= index) AND (parent_id); j--) {
            if (list[j].ObjectID IS parent_id) {
               if (!(list[j].Flags & RNF_VISIBLE)) {
                  skip = TRUE;
                  break;
               }

               ClipRectangle(&childexpose, (struct ClipRectangle *)&list[j].Left);

               parent_id = list[j].ParentID;
            }
         }
         if (skip) continue;

         /*** Skip this drawable if there is nothing to be seen (lies outside the expose boundary) ***/

         if ((childexpose.right <= childexpose.left) OR (childexpose.bottom <= childexpose.top)) continue;
      }

      /*** Do the expose ***/

      if (AccessObject(list[i].BitmapID, glMaxBitmapWait, &bitmap) IS ERR_Okay) {
         if (!video) {
            if ((video = AccessVideo(list[index].ScreenID))) {
                video->XOffset = 0;
                video->YOffset = 0;
            }
            else DPrintF("@ExposeDrawable:","Unable to access screen #%d.", list[index].ScreenID);
         }

         if (video) ExposeBuffer(list, i, i, childexpose.left, childexpose.top, childexpose.right, childexpose.bottom, video, bitmap);

         ReleaseObject(bitmap);
      }
      else {
         #ifdef DBG_DRAW
            DPrintF("ExposeDrawable:","Unable to access internal bitmap, sending delayed expose message.");
         #endif

         expose.XCoord = childexpose.left   - list[i].AbsX;
         expose.YCoord = childexpose.top    - list[i].AbsY;
         expose.Width  = childexpose.right  - childexpose.left;
         expose.Height = childexpose.bottom - childexpose.top;
         expose.Flags  = NULL;
         DelayMsg(MT_Expose, list[i].ObjectID, &expose);
      }
   }

   /* These flags should be set if the drawable has had some area of it
   ** redrawn prior to the ExposeDrawable() call.  This can be very important if the
   ** application has been writing to the drawable directly rather than the more
   ** conventional drawing procedures.
   **
   ** If the drawable bitmap has not been changed, volatile redrawing just wastes
   ** CPU time for the user.
   */

   if (Flags & (EXF_REDRAW_VOLATILE|EXF_REDRAW_VOLATILE_OVERLAP)) {
      /* Redraw any volatile regions that intersect our expose area (such regions must be
      ** updated to reflect the new background graphics).  Note that this routine does a
      ** fairly deep scan, due to the selective area copying features in our system
      ** (i.e. we cannot just skim over the stuff that is immediately in front of us).
      **
      ** EXF_REDRAW_VOLATILE: Redraws every single volatile object that intersects the expose,
      **    including internal volatile children.
      **
      ** EXF_REDRAW_VOLATILE_OVERLAP: Only redraws volatile objects that obscure the expose
      **    from a position outside of the drawable and its children.  Useful if no redrawing
      **    has occurred internally, but the render object has been moved to a new position
      **    and the parents need to be redrawn.
      */

      level = list[index].Level + 1;

      if ((Flags & EXF_REDRAW_VOLATILE_OVERLAP) ){ //OR (Flags & EXF_CHILDREN)) {
         /* All children in our area have already been redrawn or do not need
         ** redrawing, so skip past them.
         */
         for (i=index+1; list[i].Level > list[index].Level; i++);
         if (list[i-1].Flags & RNF_CURSOR) i--; /* Never skip past the cursor */
      }
      else i = index + 1;

      #ifdef DBG_DRAW
         DPrintF("~ExposeDrawable:","Redraw volatiles from idx %d, area %dx%d,%dx%d", i, abs.left, abs.top, abs.right - abs.left, abs.bottom - abs.top);
      #endif

      if (i < glVolatileIndex) i = glVolatileIndex; /* Volatile index allows the starting point to be specified */

      for (; (list[i].Level >= 1); i++) {
         if (list[i].Level < level) level = list[i].Level; /* Drop the comparison level down */

         if (list[i].Level <= level) {
            if ((list[i].Flags & (RNF_VOLATILE|RNF_CURSOR)) AND (list[i].Flags & RNF_VISIBLE)) {
               if (list[i].ObjectID IS RenderID) continue;

               if ((list[i].Right > abs.left) AND (list[i].Bottom > abs.top) AND
                   (list[i].Left < abs.right) AND (list[i].Top < abs.bottom)) {

                  if (list[i].Flags & RNF_POSTCOMPOSITE); /* Post-composites never require redrawing because they are not completely volatile */
                  else RedrawDrawable(list[i].ObjectID, abs.left, abs.top, abs.right, abs.bottom, NULL);

                  ExposeDrawable(list[i].ObjectID, abs.left, abs.top, abs.right, abs.bottom, EXF_ABSOLUTE);
               }
            }
         }
      }

      #ifdef DBG_DRAW
         StepBack();
      #endif
   }
   else {
      /* Look for a software cursor at the end of the renderlist and redraw it.  (We have
      ** to redraw the cursor as ExposeBuffer() ignores it for optimisation purposes.)
      */

      #ifdef DBG_DRAW
         DPrintF("~ExposeDrawable:","Exposing cursor.");
      #endif

      for (i=index; list[i].ObjectID; i++);
      i--;

      if ((list[i].Flags & RNF_CURSOR) AND (list[i].ObjectID != RenderID)) {
         if ((list[i].Right > abs.left) AND (list[i].Bottom > abs.top) AND
             (list[i].Left < abs.right) AND (list[i].Top < abs.bottom)) {

            if (list[i].Flags & RNF_POSTCOMPOSITE); /* Post-composites never require redrawing because they are not completely volatile */
            else RedrawDrawable(list[i].ObjectID, abs.left, abs.top, abs.right, abs.bottom, NULL);

            ExposeDrawable(list[i].ObjectID, abs.left, abs.top, abs.right, abs.bottom, EXF_ABSOLUTE);
         }
      }

      #ifdef DBG_DRAW
         StepBack();
      #endif
   }

exit:
   if (list) ReleaseRenderList(ARF_READ);
   if (video) ReleaseVideo();

   #ifdef DBG_DRAW
      StepBack();
   #endif

   return(ERR_Okay);
}

/****************************************************************************/

static void ExposeBuffer(struct RenderList *list, WORD Index, WORD ScanIndex, LONG Left, LONG Top,
                         LONG Right, LONG Bottom, struct Bitmap *Video,
                         struct Bitmap *Bitmap)
{
   WORD i, j, owner;
   LONG listx, listy, listright, listbottom;

   /*** Scan for overlapping parent/sibling regions and avoid them ***/

   for (i=ScanIndex+1; list[i].Level > 1; i++) {

      /*** Skip past non-visible areas and their content ***/

      if (!(list[i].Flags & RNF_VISIBLE)) {
         j = list[i].Level;
         while (list[i+1].Level > j) i++;
         continue;
      }

      if (list[i].Flags & (RNF_REGION|RNF_CURSOR)) goto skip; /* Ignore regions and the cursor */

      if (list[i].BitmapID IS list[Index].BitmapID) continue; /* Ignore any children that form part of our bitmap space */

      listx      = list[i].Left;
      listy      = list[i].Top;
      listright  = list[i].Right;
      listbottom = list[i].Bottom;

      if ((listx < Right) AND (listy < Bottom) AND (listright > Left) AND (listbottom > Top)) {

         if (listx <= Left) listx = Left;
         else ExposeBuffer(list, Index, ScanIndex, Left, Top, listx, Bottom, Video, Bitmap); /* left */

         if (listright >= Right) listright = Right;
         else ExposeBuffer(list, Index, ScanIndex, listright, Top, Right, Bottom, Video, Bitmap); /* right */

         if (listy <= Top) listy = Top;
         else ExposeBuffer(list, Index, ScanIndex, listx, Top, listright, listy, Video, Bitmap); /* top */

         if (listbottom < Bottom) ExposeBuffer(list, Index, ScanIndex, listx, listbottom, listright, Bottom, Video, Bitmap); /* bottom */

         if (list[i].Flags & RNF_INVISIBLE) {
            /* In the case of invisible regions, we will have split the expose process as normal.  However,
            ** we also need to look deeper into the invisible region to discover if there is more that
            ** we can draw, depending on the content of the invisible region.
            */

            listx      = list[i].Left;
            listy      = list[i].Top;
            listright  = list[i].Right;
            listbottom = list[i].Bottom;

            if (Left > listx) listx = Left;
            if (Top > listy) listy = Top;
            if (Right < listright) listright = Right;
            if (Bottom < listbottom) listbottom = Bottom;

            ExposeBuffer(list, Index, i, listx, listy, listright, listbottom, Video, Bitmap);
         }

         return;
      }

      /* Skip past any children of the overlapping object.  This ensures that
      ** we only look at immediate parents and siblings that are in our way.
      */
skip:
      j = i + 1;
      while (list[j].Level > list[i].Level) j++;
      i = j - 1;
   }

   #ifdef DBG_DRAW
      DPrintF("~ExposeBuffer()","[%d] %dx%d,%dx%d Bmp: %d, Idx: %d/%d", list[Index].ObjectID, Left, Top, Right - Left, Bottom - Top, list[Index].BitmapID, Index, ScanIndex);
   #endif

   /*** The region is not obscured, so perform the redraw ***/

   owner = FindBitmapOwner(list, Index);

   /* Turn off offsets and set the clipping to match the source bitmap exactly
   ** (i.e. nothing fancy happening here)
   */

   Bitmap->XOffset = 0;
   Bitmap->YOffset = 0;

   Bitmap->ClipLeft   = 0;
   Bitmap->ClipTop    = 0;
   Bitmap->ClipRight  = list[Index].Width;
   Bitmap->ClipBottom = list[Index].Height;
   if (Bitmap->ClipRight  > Bitmap->Width)  Bitmap->ClipRight  = Bitmap->Width;
   if (Bitmap->ClipBottom > Bitmap->Height) Bitmap->ClipBottom = Bitmap->Height;

   /*** Set the clipping so that we are only drawing to the video display area that has been exposed ***/

   Video->ClipLeft   = Left;
   Video->ClipTop    = Top;
   Video->ClipRight  = Right;
   Video->ClipBottom = Bottom;
   if (Video->ClipLeft < 0) Video->ClipLeft = 0;
   if (Video->ClipTop  < 0) Video->ClipTop  = 0;
   if (Video->ClipRight  > Video->Width)  Video->ClipRight  = Video->Width;
   if (Video->ClipBottom > Video->Height) Video->ClipBottom = Video->Height;

   /* If POSTCOMPOSITE is in use, this means we have to do compositing on
   ** the fly.  This involves copying the background graphics into a temporary
   ** buffer, then blitting the composite buffer to the display.
   */

   if ((list[Index].Flags & RNF_VOLATILE) AND (list[Index].Flags & RNF_POSTCOMPOSITE)) {
      struct ClipRectangle clip;

      if (glComposite) {
         if (glComposite->BitsPerPixel != Video->BitsPerPixel) {
            acFree(glComposite);
            glComposite = NULL;
         }
         else if ((glComposite->Width < list[Index].Width) OR (glComposite->Height < list[Index].Height)) {
            acResize(glComposite, (list[Index].Width > glComposite->Width) ? list[Index].Width : glComposite->Width,
                                  (list[Index].Height > glComposite->Height) ? list[Index].Height : glComposite->Height,
                                  0);
         }
      }

      if (!glComposite) {
         if (CreateObject(ID_BITMAP, NF_UNTRACKED, &glComposite, NULL,
            FID_Width|TLONG,  list[Index].Width,
            FID_Height|TLONG, list[Index].Height,
            TAGEND) != ERR_Okay) goto exit;

         SetOwner(glComposite, modRender);
      }

      /*** Build the background in our buffer ***/

      clip.Left   = Left;
      clip.Top    = Top;
      clip.Right  = Right;
      clip.Bottom = Bottom;
      PrepareBackground(NULL, list, Index, glComposite, &clip, STAGE_POSTCOMPOSITE);

      /*** Blend the drawable's graphics into the composited buffer ***/

      Bitmap->Flags |= BMF_ALPHABLEND;
      ActionTags(MT_BlitArea, Bitmap, glComposite, NULL, 0, 0, list[Index].Width, list[Index].Height, 0, 0);
      Bitmap->Flags &= ~BMF_ALPHABLEND;

      /*** Copy the fully composited buffer to the display ***/

      ActionTags(MT_BlitArea, glComposite, Video, NULL,
         0, 0,
         list[Index].Width, list[Index].Height,
         list[Index].Left, list[Index].Top);
   }
   else {
      /*** Copy the buffered graphic to the video display ***/

      ActionTags(MT_BlitArea, Bitmap, Video, NULL,
         list[Index].Left-list[owner].Left, list[Index].Top-list[owner].Top, /* Src X/Y (bitmap relative) */
         list[Index].Width, list[Index].Height,
         list[Index].Left, list[Index].Top);                                 /* Dest X/Y (absolute screen position) */
   }

   #ifdef DBG_DRAW
     if (glDebugExposure) {
        ActionTags(MT_DrawRectangle, Video, Video->ClipLeft, Video->ClipTop, Video->ClipRight - Video->ClipLeft, Video->ClipBottom - Video->ClipTop,
           PackPixel(Video, RandomNumber(255), RandomNumber(255), RandomNumber(255)), FALSE);
     }
   #endif

exit:
   #ifdef DBG_DRAW
      StepBack();
   #endif
   return;
}

/*****************************************************************************

###FUNCTION###
Name:     ForbidExpose()
Short:    Temporarily turns off display exposures.
Synopsis: void ForbidExpose(void)

###DESCRIPTION###
<p>The ForbidExpose() function turns off all future display exposures
attempted by the calling process.  This effectively means that all calls to
ExposeDrawable() will return immediately (with ERR_Okay status).</p>

<p>Turning off display exposures is often useful when performing a number of
tasks (such as resizing many small drawables) that will result in multiple
exposures.  This may potentially cause display flickering that could be optimised
out of the system if the exposures were covered manually with a single
ExposeDrawable() call.  Using ForbidExpose() to turn off the exposes will
allow you to achieve such an optimisation.</p>

<p>This function call must be followed up with a call to PermitExpose() at
some stage if the graphics system is to return to its normal state of operation.</p>

<p>Calls to ForbidExpose() will nest.</p>

###SEE ALSO###
Function: PermitExpose()

###END###

*****************************************************************************/

static void ForbidExpose(void)
{
   glNoExpose++;
}

/*****************************************************************************

###FUNCTION###
Name:     PermitExpose()
Short:    Undoes the the ForbidExpose() function.
Synopsis: void PermitExpose(void)

###DESCRIPTION###
<p>The PermitExpose() function reverse the effect of the ForbidExpose() function.
Please read the documentation for ForbidExpose() for more information.</p>

###SEE ALSO###
Function: PermitExpose()

###END###

*****************************************************************************/

static void PermitExpose(void)
{
   glNoExpose--;
}

/*****************************************************************************

###FUNCTION###
Name:     GetRenderInfo
Short:    Retrieves display information for any render object without having to access it directly.
Synopsis: ERROR GetRenderInfo(OBJECTID RenderID, struct RenderInfo *Info, LONG InfoSize)

###DESCRIPTION###
<p>The GetRenderInfo function is used for quickly retrieving basic information on render
objects without having to access them through the slower AccessObject() function.  You
need to supply this function with the ID of the render object that you want to
analyse and provide a return structure for the resulting feedback.</p>

<p>The RenderInfo structure contains the following fields of information:</p>

<pre>
   OBJECTID ParentID   The object that contains the render area.
   OBJECTID BitmapID   Drawable bitmap buffer.
   OBJECTID FocusID    Indicates the focus route.
   OBJECTID ScreenID   If set, indicates the screen owned by the drawable.
   MEMORYID DataMID    Memory ID that refers to the bitmap data.
   LONG Flags          Render flags (RNF_VISIBLE etc).
   LONG XCoord         Horizontal coordianate.
   LONG YCoord         Vertical coordinate.
   LONG Width          The width of the drawable area.
   LONG Height         The height of the drawable area.
   LONG AbsX           Absolute X coordinate.
   LONG AbsY           Absolute Y coordinate.
   WORD Level          The drawable's level within the render tree.
   WORD TopLimit       Top limit (for restricting movement).
   WORD LeftLimit      Left limit.
   WORD RightLimit     Right limit.
   WORD BottomLimit    Bottom limit.
   LONG LineWidth      The bytewidth of each line in the bitmap data.
   BYTE BytesPerPixel  The number of bytes per pixel in the bitmap data.
   BYTE BitsPerPixel   The number of bits per pixel in the bitmap data.
</pre>

###INPUT###
RenderID: The unqiue ID of the render object that you want information on.
Info: Pointer to a RenderInfo structure that will be filled out by this function.
InfoSize: Must be set to 'sizeof(struct RenderInfo)'.

###RESULT###
ERR_Okay: The render information was successfully retrieved.
ERR_Args: Invalid arguments were specified.
ERR_Search: The supplied RenderID did not refer to a recognised render object.
ERR_AccessMemory: Failed to access the internal renderlist memory structure.

###END###

*****************************************************************************/

static ERROR GetRenderInfo(OBJECTID RenderID, struct RenderInfo *Info, LONG InfoSize)
{
   struct RenderList *list;
   WORD i, root;

   if ((!RenderID) OR (!Info) OR (InfoSize <= 0)) return(SendError(ERH_GetRenderInfo, ERR_Args));

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      if ((i = FindDrawableIndex(list, RenderID)) IS -1) {
         ReleaseRenderList(ARF_READ);
         return(ERR_Search);
      }

      root = FindBitmapOwner(list, i);

      if (InfoSize IS sizeof(struct RenderInfoV1)) {
         /* The earlier version of this structure uses 16-bit dimension fields */
         ((struct RenderInfoV1 *)Info)->ParentID      = list[i].ParentID;
         ((struct RenderInfoV1 *)Info)->BitmapID      = list[i].BitmapID;
         ((struct RenderInfoV1 *)Info)->FocusID       = list[i].FocusID;
         ((struct RenderInfoV1 *)Info)->ScreenID      = list[i].ScreenID;
         ((struct RenderInfoV1 *)Info)->DataMID       = list[root].DataMID;
         ((struct RenderInfoV1 *)Info)->Flags         = list[i].Flags;
         ((struct RenderInfoV1 *)Info)->XCoord        = list[i].XCoord;
         ((struct RenderInfoV1 *)Info)->YCoord        = list[i].YCoord;
         ((struct RenderInfoV1 *)Info)->Width         = list[i].Width;
         ((struct RenderInfoV1 *)Info)->Height        = list[i].Height;
         ((struct RenderInfoV1 *)Info)->AbsX          = list[i].AbsX;
         ((struct RenderInfoV1 *)Info)->AbsY          = list[i].AbsY;
         ((struct RenderInfoV1 *)Info)->Level         = list[i].Level;
         ((struct RenderInfoV1 *)Info)->LeftLimit     = list[i].LeftLimit;
         ((struct RenderInfoV1 *)Info)->TopLimit      = list[i].TopLimit;
         ((struct RenderInfoV1 *)Info)->RightLimit    = list[i].RightLimit;
         ((struct RenderInfoV1 *)Info)->BottomLimit   = list[i].BottomLimit;
         ((struct RenderInfoV1 *)Info)->BytesPerPixel = list[root].BytesPerPixel;
         ((struct RenderInfoV1 *)Info)->BitsPerPixel  = list[root].BitsPerPixel;
         ((struct RenderInfoV1 *)Info)->LineWidth     = list[root].LineWidth;
      }
      else {
         Info->ParentID = list[i].ParentID;
         Info->BitmapID = list[i].BitmapID;
         Info->FocusID  = list[i].FocusID;
         Info->ScreenID = list[i].ScreenID;
         Info->DataMID  = list[root].DataMID;
         Info->Flags    = list[i].Flags;
         Info->XCoord   = list[i].XCoord;
         Info->YCoord   = list[i].YCoord;
         Info->Width    = list[i].Width;
         Info->Height   = list[i].Height;
         Info->AbsX     = list[i].Left;
         Info->AbsY     = list[i].Top;
         Info->Level    = list[i].Level;
         Info->LeftLimit     = list[i].LeftLimit;
         Info->TopLimit      = list[i].TopLimit;
         Info->RightLimit    = list[i].RightLimit;
         Info->BottomLimit   = list[i].BottomLimit;
         Info->BytesPerPixel = list[root].BytesPerPixel;
         Info->BitsPerPixel  = list[root].BitsPerPixel;
         Info->LineWidth     = list[root].LineWidth;
      }

      ReleaseRenderList(ARF_READ);
      return(ERR_Okay);
   }
   else return(SendError(ERH_GetRenderInfo, ERR_AccessMemory));
}

/*****************************************************************************

###FUNCTION###
Name:     GetUserFocus
Short:    Returns the ID of the drawable that currently has the user's focus.
Synopsis: OBJECTID GetUserFocus(void)

###DESCRIPTION###
<p>This function returns the unique ID of the drawable that has the user's
focus.</p>

###RESULT###
<p>Returns the ID of the render object that has the user focus, or NULL on
failure.</p>

###END###

*****************************************************************************/

static OBJECTID GetUserFocus(void)
{
   OBJECTID *focuslist, objectid;

   if (AccessMemory(RPM_FocusList, MEM_READ, &focuslist) IS ERR_Okay) {
      objectid = focuslist[0];
      ReleaseMemory(focuslist);
      return(objectid);
   }
   else return(NULL);
}

/*****************************************************************************

###FUNCTION###
Name:     LockDrawableBitmap
Short:    Returns a bitmap that represents the video area covered by the render object.
Synopsis: ERROR LockDrawableBitmap(OBJECTID RenderID, struct Bitmap **Bitmap, LONG *Info)

###DESCRIPTION###
<p>Use the LockDrawableBitmap() function when you would like direct access
to the bitmap information of a render object.  Full access to the graphics data
will be granted through the Bitmap object that is returned by this function.  As the
rendering buffer will be locked from the system whilst you retain access,
you must keep your access time to an absolute minimum or desktop performance
may suffer.</p>

<p>Repeated calls to LockDrawableBitmap() will nest.  To release a drawable
bitmap, call the UnlockDrawableBitmap() function.</p>

<p>If you require long-term access to the graphics data, consider using
the AccessDrawableSurface() function instead.</p>

###INPUT###
RenderID: Object ID of the render object that you want to lock.
Bitmap: The resulting bitmap will be returned in this parameter.
Info: Special flags may be returned in this parameter.  If LVF_EXPOSECHANGES is set, you must expose your changes if you want to see them on the display.

###RESULT###
ERR_Okay: The drawable area was successfully locked.
ERR_Args: Invalid arguments were specified.

###SEE ALSO###
Function: AccessDrawableSurface() UnlockDrawableBitmap()

###END###

*****************************************************************************/

static ERROR LockDrawableBitmap(OBJECTID RenderID, struct Bitmap **Bitmap, LONG *Info)
{
   struct RenderList *list;
   LONG i, j;

   if (Info) *Info = NULL;

   if ((!RenderID) OR (!Bitmap)) return(SendError(ERH_LockBitmap, ERR_Args));

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS RenderID) {

            if (!list[i].BitmapID) {
               DPrintF("@LockDrawableBitmap:","No Bitmap is logged against object #%d.", RenderID);
               ReleaseRenderList(ARF_READ);
               return(ERR_Failed);
            }

            /* Gain access to the bitmap buffer and set the clipping
            ** and offsets to the correct values.
            */

            if (AccessObject(list[i].BitmapID, 5000, Bitmap) IS ERR_Okay) {

               for (j=i; (j > 0) AND (list[j-1].BitmapID IS list[i].BitmapID); j--);

               Bitmap[0]->XOffset    = list[i].AbsX - list[j].AbsX;
               Bitmap[0]->YOffset    = list[i].AbsY - list[j].AbsY;
               Bitmap[0]->ClipLeft   = 0;
               Bitmap[0]->ClipTop    = 0;
               Bitmap[0]->ClipRight  = list[i].Width;
               Bitmap[0]->ClipBottom = list[i].Height;

               if (Bitmap[0]->ClipRight + Bitmap[0]->XOffset > Bitmap[0]->Width){
                  Bitmap[0]->ClipRight = Bitmap[0]->Width - Bitmap[0]->XOffset;

                  if (Bitmap[0]->ClipRight < 0) {
                     ReleaseRenderList(ARF_READ);
                     return(ERR_Failed);
                  }
               }

               if (Bitmap[0]->ClipBottom + Bitmap[0]->YOffset > Bitmap[0]->Height){
                  Bitmap[0]->ClipBottom = Bitmap[0]->Height - Bitmap[0]->YOffset;

                  if (Bitmap[0]->ClipBottom < 0) {
                     ReleaseRenderList(ARF_READ);
                     return(ERR_Failed);
                  }
               }
            }
            else {
               ReleaseRenderList(ARF_READ);
               return(SendError(ERH_LockBitmap, ERR_AccessObject));
            }

            if (Info) {
               /* The developer will have to send an expose signal - unless
               ** the exposure can be gained for 'free' (possible if the Draw
               ** action has been called on the Render object).
               */

               if (glFreeExpose IS list[i].BitmapID);
               else *Info |= LVF_EXPOSECHANGES;
            }

            ReleaseRenderList(ARF_READ);
            return(ERR_Okay);
         }
      }

      ReleaseRenderList(ARF_READ);
      return(SendError(ERH_LockBitmap, ERR_Search));
   }
   else return(SendError(ERH_LockBitmap, ERR_AccessMemory));
}

/*****************************************************************************

###FUNCTION###
Name:     ReleaseDrawableSurface
Short:    Releases any earlier call to AccessDrawableSurface().
Synopsis: ERROR ReleaseDrawableSurface(OBJECTID RenderID, struct BitmapSurface *Surface)

###DESCRIPTION###
<p>This function will release bitmap data that has been acquired from previous
calls to AccessDrawableSurface().</p>

###INPUT###
RenderID: Reference to the render object that you are releasing.
Surface: Pointer to the BitmapSurface structure being released.

###RESULT###
ERR_Okay: Release successful.
ERR_Args: Invalid arguments were supplied.

###SEE ALSO###
Function: AccessDrawableSurface

###END###

*****************************************************************************/

static ERROR ReleaseDrawableSurface(OBJECTID RenderID, struct BitmapSurface *Surface)
{
   if (!Surface) return(SendError(ERH_ReleaseSurface, ERR_Args));

   if (Surface->Version >= 2) {
      if (Surface->Private) ReleaseObject(Surface->Private);
      else ReleaseMemory(Surface->Data);
   }
   else ReleaseMemory(Surface->Data);

   return(ERR_Okay);
}

/*****************************************************************************

###PRIVATE###
Name:     ReleaseRenderList()
Short:    Releases access to the internal renderlist array.
Synopsis: void ReleaseRenderList(LONG Flags)

###DESCRIPTION###

###INPUT###
Flags: You must set the same flags as in your original call to AccessRenderList().

###SEE ALSO###
Function: AccessRenderList

###END###

*****************************************************************************/

static void ReleaseRenderList(LONG Flags)
{
#ifdef SEMAPHORES
#ifdef SAFE_SEMAPHORES
   Flags = ARF_WRITE;
#endif

   if (Flags & (ARF_WRITE|ARF_UPDATE)) ReleaseSemaphore(glListSemaphore, NULL);
   else if (Flags & ARF_READ) ReleaseSemaphore(glListSemaphore, SMF_NONBLOCKING);
   else DPrintF("@ReleaseRenderList()","ARF_READ or ARF_WRITE must be specified.");

#else

   if (glListCount > 0) {
      glListCount--;
      if (!glListCount) {
         ReleaseMemory(glRenderList);
         glRenderList = NULL;
      }
   }
   else DPrintF("@ReleaseRenderList()","Function called without an existing lock.");

#endif
}

/*****************************************************************************

###FUNCTION###
Name:     SetDrawableRegion()
Short:    Writes clipping information for a rendered area to a region object.
Synopsis: ERROR SetDrawableRegion(OBJECTID RenderID, *Region)

###DESCRIPTION###
<p>This function is used in conjunction with the Region class to define the
available drawing space within a rendered object.  This function will scan
the content of the render object referred to by RenderID, then produces
a list of obscuring regions and sets those against the Region object.
Overlapping parent regions that intersect with the buffered drawing space
will also be included in the region list.</p>

<p>This function is intended for use after calling the LockDrawableBitmap()
function, as this is the only way to get direct access to the surface of a
rendered object.  You will need to use the region information in conjunction
with a Pen object in order to draw clipped graphics to the bitmap surface.</p>

###INPUT###
RenderID: The ID of the render object that you want to scan.
Region: Pointer to a region object that will receive the clipping information.

###SEE ALSO###
Function: LockDrawableBitmap()

###END###

*****************************************************************************/

static ERROR SetDrawableRegion(OBJECTID RenderID, OBJECTPTR Region)
{
   struct mtAddRegionRect add;
   struct RenderList *list;
   WORD i, root, index;

   if ((!RenderID) OR (!Region)) return(SendError(ERH_SetRegion, ERR_Args));

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      /*** Find the render object in the render list ***/

      for (index=0; (list[index].ObjectID) AND (list[index].ObjectID != RenderID); index++);
      if (list[index].ObjectID != RenderID) {
         ReleaseRenderList(ARF_READ);
         return(ERR_Search);
      }

      /*** Clear region information from the region object ***/

      acClear(Region);

      /*** Scan for immediate children and pass them to the region object ***/

      root = FindBitmapOwner(list, index);
      for (i=index+1; (list[i].Level >= list[index].Level); i++) {

         if (list[i].Level IS (list[index].Level + 1)) {
            add.XCoord = list[i].XCoord;
            add.YCoord = list[i].YCoord;
            add.Width  = list[i].Width;
            add.Height = list[i].Height;
            //add.Flags  = APF_OBSCURING;
            Action(MT_AddRegionRect, Region, &add);
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ERR_Okay);
   }
   else return(SendError(ERH_SetRegion, ERR_AccessMemory));

   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     UnlockDrawableBitmap
Short:    Unlocks any earlier call to LockDrawableBitmap().
Synopsis: ERROR UnlockDrawableBitmap(OBJECTID RenderID, struct Bitmap *Bitmap)

###DESCRIPTION###
<p>Call the UnlockDrawableBitmap() function to release a render object from
earlier calls to LockDrawableBitmap().</p>

###INPUT###
RenderID: The ID of the render object that you are releasing.
Bitmap: Pointer to the Bitmap structure returned earlier by LockDrawableBitmap().

###RESULT###
ERR_Okay: The bitmap has been unlocked successfully.
ERR_Args: Invalid arguments were specified.

###SEE ALSO###
Function: LockDrawableBitmap()

###END###

*****************************************************************************/

static ERROR UnlockDrawableBitmap(OBJECTID RenderID, struct Bitmap *Bitmap)
{
   if ((!RenderID) OR (!Bitmap)) return(SendError(ERH_UnlockBitmap, ERR_Args));

   ReleaseObject(Bitmap);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: FindBitmapOwner()
** Short:    Scans the renderlist for the 'true owner' of a given bitmap.
*/

static WORD FindBitmapOwner(struct RenderList *List, WORD Index)
{
   WORD owner, i;

   owner = Index;
   for (i=Index; i >= 0; i--) {
      if (List[i].ObjectID IS List[owner].ParentID) {
         if (List[i].BitmapID != List[owner].BitmapID) return(owner);
         owner = i;
      }
   }
   return(owner);
}

/*****************************************************************************
** Internal: RenderObscured()
** Short:    Determines if a render object is obscured by other drawable objects.
*/
#if 0
static LONG RenderObscured(struct Render *Self, struct RenderList *list, WORD Index, LONG AbsX, LONG AbsY)
{
   WORD i, level, vislevel;
   LONG right, bottom;

   level  = list[Index].Level;
   i      = Index + 1;
   right  = AbsX + Self->Width;
   bottom = AbsY + Self->Height;
   while (list[i].ObjectID) {
      if ((list[i].Level <= level) AND (i != Index) AND (list[i].Flags & RNF_VISIBLE)) {
         if ((list[i].AbsX < right) AND (list[i].AbsY < bottom) AND
             (list[i].Right > AbsX) AND (list[i].Bottom > AbsY)) {
            return(TRUE);
         }
      }

      if (!(list[i].Flags & RNF_VISIBLE)) {
         /*** Skip this render object's children as they are invisible ***/
         vislevel = list[i++].Level;
         while (list[i].Level > vislevel) i++;
      }
      else i++;
   }

   return(FALSE);
}
#endif
/*****************************************************************************
** Function: TrackLayer()
**
** This function is responsible for inserting new render objects into the
** list of layers for positional/depth management.
**
** Drawable levels start at 1, which indicates the top-most level.
*/

static ERROR TrackLayer(struct Render *Self)
{
   struct RenderList *list;
   WORD absx, absy, level, i, j;

   if ((list = AccessRenderList(ARF_WRITE, NULL))) {

      /*** Find the position at which the render object should be inserted ***/

      absx = 0;
      absy = 0;
      if (!Self->ParentID) {
         /*** Insert the render object at the end of the list ***/

         for (i=0; list[i].ObjectID; i++);
         if (i >= SIZE_RENDERLIST) goto overflow;
         level = 1;
      }
      else {
         /* Go to the end of the object list and at the same time make a
         ** note of the level at which we will lie.
         */

         level = 0;
         for (i=0; (list[i].ObjectID); i++) {
            if (list[i].ObjectID IS Self->ParentID) {
               level = list[i].Level + 1;
               absx  = list[i].AbsX + Self->XCoord;
               absy  = list[i].AbsY + Self->YCoord;
            }
         }

         /* Step back through the list to find the first child of our
         ** parent, or the parent itself.
         */

         if (Self->Flags & RNF_STICKTOFRONT) {
            while ((i > 0) AND (list[i-1].ObjectID != Self->ParentID) AND ((list[i-1].ParentID != Self->ParentID) OR ((list[i-1].Flags & RNF_POINTER)))) i--;
         }
         else {
            while ((i > 0) AND (list[i-1].ObjectID != Self->ParentID) AND ((list[i-1].ParentID != Self->ParentID) OR ((list[i-1].Flags & RNF_STICKTOFRONT)))) i--;
         }

         while ((list[i].ObjectID) AND (list[i].Level > level)) i++;

         if ((i <= 0) OR (!level)) {
            ReleaseRenderList(ARF_WRITE);
            DPrintF("@TrackLayer:","[Render:%d] Failed to find registration for parent object #%d.", Self->Head.UniqueID, Self->ParentID);
            return(ERR_Search);
         }

         /*** Make space for insertion ***/

         if (list[i].ObjectID) {
            for (j=i; list[j].ObjectID; j++);
            CopyMemory(list+i, list+i+1, sizeof(struct RenderList) * (j-i));
         }
      }

      list[i].ParentID = Self->ParentID;
      list[i].ObjectID = Self->Head.UniqueID;
      list[i].BitmapID = Self->BufferID;
      list[i].FocusID  = Self->RouteFocusID;
      list[i].ScreenID = Self->ScreenID;
      list[i].Flags    = Self->Flags;
      list[i].XCoord   = Self->XCoord;
      list[i].YCoord   = Self->YCoord;
      list[i].AbsX     = absx;
      list[i].AbsY     = absy;
      list[i].Width    = Self->Width;
      list[i].Height   = Self->Height;
      list[i].Right    = absx + Self->Width;
      list[i].Bottom   = absy + Self->Height;
      list[i].Level    = level;
      list[i].Opacity  = Self->prvOpacity;
      list[i].LeftLimit     = Self->LeftLimit;
      list[i].TopLimit      = Self->TopLimit;
      list[i].RightLimit    = Self->RightLimit;
      list[i].BottomLimit   = Self->BottomLimit;
      list[i].BitsPerPixel  = Self->prvBitsPerPixel;
      list[i].BytesPerPixel = Self->prvBytesPerPixel;
      list[i].LineWidth     = Self->prvLineWidth;
      list[i].DataMID       = Self->prvDataMID;
      list[i].CursorID      = Self->CursorID;

      ReleaseRenderList(ARF_WRITE);
      return(ERR_Okay);
   }
   else {
      DPrintF("@TrackLayer:","[Render:%d] Failed to access the renderlist.", Self->Head.UniqueID);
      return(ERR_AccessSemaphore);
   }

overflow:
   if (list) ReleaseRenderList(ARF_WRITE);
   return(ObjectError(Self, ERH_Init, ERR_BufferOverflow));
}

/*****************************************************************************
** Function: UntrackLayer()
*/

static void UntrackLayer(OBJECTID ObjectID)
{
   struct RenderList *list;
   LONG i, end, total;

   if ((list = AccessRenderList(ARF_WRITE, NULL))) {
      for (i=0; (list[i].ObjectID) AND (list[i].ObjectID != ObjectID); i++);

      if (list[i].ObjectID IS ObjectID) {
         for (total=i+1; list[total].Level > list[i].Level; total++) {
            list[total].Flags &= ~RNF_VISIBLE;
         }

         /* If this object is at the end of the list, just clear the
         ** memory.  Otherwise, shift the objects in front of us
         ** down the list.
         */

         if (!list[total].ObjectID) {
            ClearMemory(list+i, sizeof(struct RenderList) * (total-i));
         }
         else {
            for (end=i; list[end].ObjectID; end++);
            CopyMemory(list+i+1, list+i, sizeof(struct RenderList) * (end-i));
         }
      }
      ReleaseRenderList(ARF_WRITE);
   }

}

/*****************************************************************************
** Function: UpdateRenderList()
*/

static ERROR UpdateRenderList(struct Render *Self)
{
   struct RenderList *list;
   LONG i, j;
   WORD level, absx, absy;

   if (!Self) return(SendError(ERH_Function, ERR_Args));
   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) return(ERR_Okay);

   if ((list = AccessRenderList(ARF_UPDATE, NULL))) {

      absx = 0;
      absy = 0;
      for (i=0; (list[i].ObjectID) AND (list[i].ObjectID != Self->Head.UniqueID); i++) {
         if (list[i].ObjectID IS Self->ParentID) {
            absx = list[i].AbsX + (WORD)Self->XCoord;
            absy = list[i].AbsY + (WORD)Self->YCoord;
         }
      }

      if (Self->Head.UniqueID IS list[i].ObjectID) {
         list[i].ParentID = Self->ParentID;
         list[i].ObjectID = Self->Head.UniqueID;
         list[i].BitmapID = Self->BufferID;
         list[i].FocusID  = Self->RouteFocusID;
         list[i].ScreenID = Self->ScreenID;
         list[i].XCoord   = Self->XCoord;
         list[i].YCoord   = Self->YCoord;
         list[i].AbsX     = absx;
         list[i].AbsY     = absy;
         list[i].Width    = Self->Width;
         list[i].Height   = Self->Height;
         list[i].Right    = absx + Self->Width;
         list[i].Bottom   = absy + Self->Height;
         list[i].Flags    = Self->Flags;
         list[i].Opacity  = Self->prvOpacity;
         list[i].LeftLimit     = Self->LeftLimit;
         list[i].TopLimit      = Self->TopLimit;
         list[i].RightLimit    = Self->RightLimit;
         list[i].BottomLimit   = Self->BottomLimit;
         list[i].BitsPerPixel  = Self->prvBitsPerPixel;
         list[i].BytesPerPixel = Self->prvBytesPerPixel;
         list[i].LineWidth     = Self->prvLineWidth;
         list[i].DataMID       = Self->prvDataMID;
         list[i].CursorID      = Self->CursorID;

         /*** Rebuild absolute coordinates of child objects ***/

         level = list[i].Level;
         i++;
         while ((list[i].ObjectID) AND (list[i].Level > level)) {
            for (j=i-1; j >= 0; j--) {
               if (list[j].ObjectID IS list[i].ParentID) {
                  list[i].Left   = list[j].Left + list[i].XCoord;
                  list[i].Top    = list[j].Top  + list[i].YCoord;
                  list[i].Right  = list[i].Left + list[i].Width;
                  list[i].Bottom = list[i].Top  + list[i].Height;
                  break;
               }
            }
            i++;
         }
      }

      ReleaseRenderList(ARF_UPDATE);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Function, ERR_AccessMemory));
}

/*****************************************************************************
** Internal: CheckVolatile()
*/

static UBYTE CheckVolatile(struct RenderList *list, WORD index)
{
   WORD i;

   //if (list[index].Flags & RNF_POSTCOMPOSITE) return(FALSE);
   if (list[index].Flags & RNF_VOLATILE) return(TRUE);

   /*** If there are children with custom root layers, that will force volatility ***/

   for (i=index+1; list[i].Level > list[index].Level; i++) {
      if ((list[i].Flags & RNF_CUSTOMROOT) AND (list[i].Flags & RNF_VOLATILE)) {
         //if (list[i].Flags & RNF_POSTCOMPOSITE) continue; /* Post-composite drawables never require redrawing and thus are not volatile in the expected sense */
         return(TRUE);
      }
   }

   return(FALSE);
}

/*****************************************************************************
** Internal: CheckVisibility()
** Short:    Checks if an object is visible, according to its visibility and
**           its parents visibility.
*/

static UBYTE CheckVisibility(struct RenderList *list, WORD index)
{
   OBJECTID scan;
   WORD i;

   scan = list[index].ObjectID;
   for (i=index; i >= 0; i--) {
      if (list[i].ObjectID IS scan) {
         if (!(list[i].Flags & RNF_VISIBLE)) return(FALSE);
         if (!(scan = list[i].ParentID)) return(TRUE);
      }
   }

   return(TRUE);
}

static void CheckBitmapBufferDepth(struct Render *Self, struct Bitmap *Bitmap)
{
   SCREENINFO info;
   struct acResize resize;

   if (Bitmap->Flags & BMF_FIXEDDEPTH) {
      /*** Don't change bitmaps marked as fixed-depth ***/
      return;
   }

   if ((GetScreenInfo(Self->ScreenID, &info, sizeof(info)) IS ERR_Okay) AND (info.BitsPerPixel != Bitmap->BitsPerPixel)) {
      DPrintF("Draw:","[%d] Updating buffer Bitmap %dx%dx%d to match new display depth of %dbpp.", Bitmap->Head.UniqueID, Bitmap->Width, Bitmap->Height, Bitmap->BitsPerPixel, info.BitsPerPixel);
      resize.Width  = Bitmap->Width;
      resize.Height = Bitmap->Height;
      resize.Depth  = info.BitsPerPixel;
      Action(AC_Resize, Bitmap, &resize);

      Self->prvLineWidth     = Bitmap->LineWidth;
      Self->prvBytesPerPixel = Bitmap->BytesPerPixel;
      Self->prvBitsPerPixel  = Bitmap->BitsPerPixel;
      Self->prvDataMID       = Bitmap->DataMID;
      UpdateRenderList(Self);
   }
}

/****************************************************************************/

#include "class/render_class.c"
